/*********************************************************************
Author:       Dale Roberts
File:         CHECKIRQ.C

Purpose:

This program checks out IRQs to see which ones may be used for
manually generating interrupts.  We take advantage of the behavior
of the edge/level setting register that causes an interrupt to
be generated on some IRQs when the edge/level mode is changed.

The registers that control the edge/level selection are 0x4d0 for the
first 8259, and 0x4d1 for the second 8259.  To check if we can use
the interrupt for our devious "manual" interrupt purposes, we
first disable the interrupts via the 8259 IMR at port address
0x21, then generate interrupts by setting the bits to 1's.
0 is edge mode, 1 is level mode.  The interrupt is generated
as long as the bit is 1.  If the bit is set back to 0,
the interrupt request is removed.

We can tell if we are running on the 486 PCI chipset if interrupts
are generated when the edge/level bits are set back to 0, rather
than when they are set to 1.  If this is detected, we report this
to the user and let them know that they need to make some changes
and re-compile.

If we read all 0xff from both registers, or if modifying the ports
at 0x4d0 and 0x4d1 do not cause any interrupts to be generated,
then we let the user know that they are out of luck.

*********************************************************************/
#define WIN32
#include <windows.h>
#include <stdio.h>
#include "intinfo.h"

int main()
{
	int imr, level;
	int active, initially_active, still_active;
	int irq_changed, irq_changed_after;
	int i, irq;

	HANDLE h;

	//
	//  Use GIVEIO.SYS to get access to I/O ports
	//
    h = CreateFile("\\\\.\\giveio", GENERIC_READ, 0, NULL,
					OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if(h == INVALID_HANDLE_VALUE) {
        printf("Couldn't access giveio device\n");
        return -1;
    }
	CloseHandle(h);

	// save IMR (interrupt mask register)
	imr = _inp(I8259_1_MASK);
	imr |= _inp(I8259_2_MASK) << 8;

	// Yes, it's possible (though highly unlikely) that the
	// IMR could change between the time we read it, and
	// the time we write over it.  But we cross our fingers
	// and hope that doesn't happen.

	// disable interrupts
	_outp(I8259_1_MASK, 0xff);

	// save original edge/level values
	level = _inp(I8259_1_EDGELVL);
	level |= _inp(I8259_2_EDGELVL) << 8;

	// set up to read IRR
	_outp(I8259_1_CMD, 0x0a);
	_outp(I8259_2_CMD, 0x0a);

	// read IRR; note which interrupts are initially active
	initially_active = _inp(I8259_1_CMD) | (_inp(I8259_2_CMD) << 8);

	// twiddle edge/level bits
	_outp(I8259_1_EDGELVL, 0xff);
	_outp(I8259_2_EDGELVL, 0xff);

	// read IRR; note which interrupts get activated
	//  due to edge/level bit
	active = _inp(I8259_1_CMD) | (_inp(I8259_2_CMD) << 8);

	// restore registers
	_outp(I8259_1_EDGELVL, level);
	_outp(I8259_2_EDGELVL, level >> 8);

	// read IRR; note which interrupts remain active after
	//  edge/level bit cleared
	still_active = _inp(I8259_1_CMD) | (_inp(I8259_2_CMD) << 8);

	// enable interrupts
	_outp(I8259_1_MASK, imr);

	// ***************************************************************
	//  Now take all of the information generated above and figure out
	// our situation.  If the IRRs remain unchanged, then we are out
	// of luck.  If IRRs change when edge/level bits are high, then we
	// are in good shape.  If IRRs change when edge/level bits are set
	// high then low, then we are on a 486 board, and GENIRQ and DOIRQ
	// will not be able to generate interrupts without some
	// modifications.

	printf("IMR 0x%04x\n", imr);

	// Figure out which IRQs were activated due to the Level
	//  bit begin set.
	irq_changed = (initially_active ^ active) & active;

	// Figure out which IRQs were activated due to the Level
	//  bit begin cleared.
	irq_changed_after = (initially_active ^ still_active)
						& still_active;

	printf("IRR: initial 0x%04x, after E/L SET 0x%04x,"
									"  after E/L CLEAR 0x%04x\n"
		   "                            change 0x%04x "
									"           change 0x%04x\n",
		initially_active, active, still_active,
		irq_changed, irq_changed_after);

	if(irq_changed == 0 && irq_changed_after == 0)
		printf("No interrupts generated."
				"  GENIRQ will not be able to\n"
				"generate overlapped interrupts.\n");

	if(irq_changed_after != 0)
		printf("Edge/Level change is latched in IRR"
							" (probably 486 chipset).\n"
				"GENIRQ will not be able to generate"
							" overlapped interrupts.\n");

	printf("\n Edge/Level Active:");

	active &= OFF_LIMITS_IRQ_MASK;

	for(i=1, irq=0; i < 0x10000; i <<= 1, ++irq)
		if(active & i)
			printf(" IRQ%d", irq);

	// must be active AND available (i.e., masked)
	active &= imr;

	printf("\nActive, AND masked:");

	for(i=1, irq=0; i < 0x10000; i <<= 1, ++irq)
		if(active & i)
			printf(" IRQ%d", irq);
}
