/*********************************************************************
Author:       Dale Roberts
File:         DOIRQ.C

Purpose:

Kernel-mode driver portion of the DOIRQ/GENIRQ team.  Together with
the GENIRQ application, this driver allows you to generate overlapped
interrupts and monitor NT's behavior in these situations.  It also
uses a serial port (initialized by GENIRQ) to generate interrupts for
timing purposes.

*********************************************************************/
#include <ntddk.h>
#include "intinfo.h"

#define DEVICE_NAME_STRING		L"doirq"

//
// This holds all the info that we use to communicate between
// the driver and the app.  It gets mapped into the app's
// address space.
//
INT_INFO	int_info;

//  Info for mapping of int_info to user space.
PMDL pmdl = 0;
PVOID UserAddress = 0;

// Used to free resources in DriverUnload().
UNICODE_STRING DriverRegistryPath;
PKINTERRUPT IntObjTable[16];

// Remember which IRQ (if any) has a priority boost.
int BoostIRQ;

/********************************************************************/
/*
 *  Direct IDT access section.
 *
 * We must modify the IDT so that interrupts that we use go through
 * an assembler routine that clears out the edge/level bit so that
 * the interrupt does not stay active.  This is not normally
 * necessary, but because we set the edge/level bits to LEVEL mode to
 * generate the interrupts, we need to clear the bits in order to
 * make the interrupts appear as EDGE interrutps.
 */

//
//  The Interrupt Descriptor Table entry and IDT processor register
// formats. Info from Intel processor manuals.
//
#pragma pack(push, 1)

typedef struct {
	unsigned offsetlo : 16;
	unsigned segment : 16;
	unsigned rsrvd  : 8;
	unsigned type : 5;
	unsigned dpl : 2;
	unsigned present : 1;
	unsigned offsethi : 16;
} IDTENT;

typedef struct {
        unsigned short  limit;
        IDTENT  *base;
} IDTREG;

#pragma pack(pop)

//
//  Table of the original interrupt vectors.  We don't replace
// all 16 interrupts.  We only replace the ones that we use.
//
ULONG OldIRQTable[16];

//
// Extern references to the assembler interrupt handlers.
//
extern ULONG asmIRQ0Handler(void);
extern ULONG asmIRQ1Handler(void);
extern ULONG asmIRQ2Handler(void);
extern ULONG asmIRQ3Handler(void);
extern ULONG asmIRQ4Handler(void);
extern ULONG asmIRQ5Handler(void);
extern ULONG asmIRQ6Handler(void);
extern ULONG asmIRQ7Handler(void);
extern ULONG asmIRQ8Handler(void);
extern ULONG asmIRQ9Handler(void);
extern ULONG asmIRQ10Handler(void);
extern ULONG asmIRQ11Handler(void);
extern ULONG asmIRQ12Handler(void);
extern ULONG asmIRQ13Handler(void);
extern ULONG asmIRQ14Handler(void);
extern ULONG asmIRQ15Handler(void);

/*********************************************************************
  Given an IRQ number, return the address of the proper assembly
language handler.
*********************************************************************/
ULONG GetHandlerAddress(int irqnum)
{
	switch(irqnum) {
		case 0: return (ULONG)asmIRQ0Handler;
		case 1: return (ULONG)asmIRQ1Handler;
		case 2: return (ULONG)asmIRQ2Handler;
		case 3: return (ULONG)asmIRQ3Handler;
		case 4: return (ULONG)asmIRQ4Handler;
		case 5: return (ULONG)asmIRQ5Handler;
		case 6: return (ULONG)asmIRQ6Handler;
		case 7: return (ULONG)asmIRQ7Handler;
		case 8: return (ULONG)asmIRQ8Handler;
		case 9: return (ULONG)asmIRQ9Handler;
		case 10: return (ULONG)asmIRQ10Handler;
		case 11: return (ULONG)asmIRQ11Handler;
		case 12: return (ULONG)asmIRQ12Handler;
		case 13: return (ULONG)asmIRQ13Handler;
		case 14: return (ULONG)asmIRQ14Handler;
		case 15: return (ULONG)asmIRQ15Handler;
	}
}

/*********************************************************************
  Poke a vector into the IDT.

 Note that the desired interrupt should be enabled before calling this
routine.  Otherwise, NT will not enable the interrupt. Apparently, if
the IDT entry is not pointing to NT's default interrupt handler, it
will not enable it when IoConnectInterrupt() is called.
*********************************************************************/
ULONG PokeIDT(int vector, ULONG Addr)
{
	IDTREG idtreg;
	ULONG OldVect;
	KIRQL OrigIrql;

	// Get Interrupt Descriptor Table pointer
	_asm sidt idtreg;

	// Disable interrupts
	KeRaiseIrql(HIGH_LEVEL, &OrigIrql);

	// Remember original vector
	OldVect = idtreg.base[vector].offsetlo |
				(idtreg.base[vector].offsethi << 16);

	// Poke in our new vector
	idtreg.base[vector].offsetlo = (ULONG)Addr & 0xffff;
	idtreg.base[vector].offsethi = (ULONG)Addr >> 16;

	// Enable interrupts
	KeLowerIrql(OrigIrql);

	return OldVect;
}

/*********************************************************************
  Poke our IRQ handler directly into the IDT.  
*********************************************************************/
void SetDirectInt(int IRQVector)
{
	KIRQL OrigIrql;

	// Disable interrupts
	KeRaiseIrql(HIGH_LEVEL, &OrigIrql);

	OldIRQTable[IRQVector-0x30] = PokeIDT(IRQVector,
								GetHandlerAddress(IRQVector-0x30));
	// Enable interrupts
	KeLowerIrql(OrigIrql);
}

/*********************************************************************
  Put the original interrupt vector back into the IDT.
*********************************************************************/
void ResetDirectInt(int IRQNum)
{
	if(OldIRQTable[IRQNum])
		PokeIDT(IRQNum+0x30, OldIRQTable[IRQNum]);
	OldIRQTable[IRQNum] = 0;
}

/*********************************************************************
  The Read Timestamp Counter "undocumented" Pentium instruction. 
Gives us the current CPU clock count in EDX:EAX.  We live on the edge
by assuming that the compiler uses EDX:EAX to return an _int64 from a
function call.
*********************************************************************/
_int64 RDTSC(void)
{
	_asm _emit 0x0f
	_asm _emit 0x31

	// EDX:EAX is returned
}

/*********************************************************************
  Unmap the int_info buffer.
*********************************************************************/
void UnMapBuffer(void)
{
	if(UserAddress != 0 && pmdl != 0) {
		MmUnmapLockedPages(UserAddress, pmdl);
		ExFreePool(pmdl);
		UserAddress = 0;
		pmdl = 0;
	}
}

/*********************************************************************
  This routine maps the int_info buffer into the GENIRQ application's
address space.
*********************************************************************/
PVOID MapBuffer(PVOID BufAdr, ULONG Length)
{
	if(pmdl != 0)
		UnMapBuffer();

	pmdl = MmCreateMdl(0, BufAdr, Length);
	MmBuildMdlForNonPagedPool(pmdl);
	UserAddress = MmMapLockedPages(pmdl, UserMode);

	// Add in offset into page.
    return (PVOID)((ULONG)UserAddress + ((ULONG)BufAdr & 0xfff));
}

/*********************************************************************
  Support routine for Trace8259().  Adds a trace entry to the table.
*********************************************************************/
void AddTrace(char *msg, int IntNum, int GenIRQMask,
				int IRR, int ISR, int IMR, char IFlag)
{
	INT_TRACE *tp;

	// Table full?
	if(int_info.TraceIdx >= NUM_TRACE_ENTRIES)
		return;

	tp = &int_info.Trace[int_info.TraceIdx];

	strncpy(tp->msg, msg, TRACE_MSG_STRING_SZ);
	tp->msg[TRACE_MSG_STRING_SZ-1] = 0;

	tp->IntNum = IntNum;
	tp->IRR = IRR;
	tp->ISR = ISR;
	tp->IMR = IMR;
	tp->IFlag = IFlag;
	tp->TimeStamp = RDTSC();
	tp->GenIRQMask = GenIRQMask;

	++int_info.TraceIdx;
}

/*********************************************************************
  Add info to trace buffer for user app to display.  We save a text
string, along with the values of the three important 8259 registers:
the IRR (Interrupt Request Register), ISR (In Service Register), and
IMR (Interrupt Mask Register).  We also save the status of the
processor's Interrupt Flag.

  Because of the sensitive nature of this routine, we must,
unfortunately, use the _asm cli, instead of KeRaiseIrql(), in order to
avoid interaction with the operating system that may affect our trace.
*********************************************************************/
void Trace8259(char *msg, int IntNum, int GenIRQMask)
{
	int IMR, IRR, ISR;
	int flags;
	KIRQL OrigIrql;

	// Disable interrupts
	_asm pushfd
	_asm cli

	// Get processor FLAGS register from stack, so we can
	//  determine value of Interrupt Flag.
	_asm mov eax, [esp]
	_asm mov flags, eax

	// Set up to read IRR
	_outp(I8259_1_CMD, 0x0a);
	_outp(I8259_2_CMD, 0x0a);
	IRR = _inp(I8259_1_CMD) | (_inp(I8259_2_CMD) << 8);

	// Set up to read ISR
	_outp(I8259_1_CMD, 0x0B);
	_outp(I8259_2_CMD, 0x0B);
	ISR = _inp(I8259_1_CMD) | (_inp(I8259_2_CMD) << 8);

	IMR = _inp(I8259_1_MASK) | (_inp(I8259_2_MASK) << 8);

	// Enable interrupts
	_asm popfd

	AddTrace(msg, IntNum, GenIRQMask,
				IRR, ISR, IMR, (char)(flags & 0x200 ? 1 : 0));
}

/*********************************************************************
  Generate one or more IRQs using the devious "twiddle the edge/level
bit" trick.  Return the actual mask used (after ANDing out any
invalid IRQs.

  Because of the sensitive nature of this routine, we must,
unfortunately, use the _asm cli, instead of KeRaiseIrql(), in order to
avoid interaction with the operating system that may affect our trace.
*********************************************************************/
int GenIRQ(int IRQMask)
{
	KIRQL OrigIrql;

	// Only set off ones that we own.
	IRQMask &= int_info.EnabledIRQMask;

	// Disable interrupts
	_asm pushfd
	_asm cli

	_outp(I8259_1_EDGELVL, _inp(I8259_1_EDGELVL) | IRQMask);
	_outp(I8259_2_EDGELVL, _inp(I8259_2_EDGELVL) | (IRQMask >> 8));

	// Enable interrupts
	_asm popfd

	return IRQMask;
}

/*********************************************************************
  Handle the serial interrupt.  Just stuff another byte into the
transmit holding register, and calculate time since we were last
called. Information is updated in the int_info structure, which is
read directly by the GENIRQ application.
*********************************************************************/
void HandleSerialIRQ(void)
{
	_int64 CurStamp;

	outp(int_info.SerPort, 0);		// load up the transmit buffer

	CurStamp = RDTSC();
	int_info.CurSerIRQInterval = (int)(CurStamp 
									- int_info.LastSerTimestamp);
	int_info.LastSerTimestamp = CurStamp;

	if(int_info.CurSerIRQInterval/int_info.CPUTicksPerUSec > 400)
		++int_info.NumDelayedInts;

	if(int_info.CurSerIRQInterval/int_info.CPUTicksPerUSec > 1000)
		++int_info.NumDelayedInts2;

	if(int_info.CurSerIRQInterval > int_info.MaxSerIRQInterval)
		int_info.MaxSerIRQInterval = int_info.CurSerIRQInterval;
}

/*********************************************************************
  Generic interrupt handler for all of the edge-generated IRQs. We are
passed the IRQ number directly in the Extension parameter (i.e.,
Extension is not actually a pointer, as intended by NT; it is an
integer). Trace entries are generated at the entry and exit of the
function so that we can monitor interrupt activity.  A trace is also
generated just when a new interrupt is activated.

  Because of the sensitive nature of this routine, we must,
unfortunately, use the _asm cli, instead of KeRaiseIrql(), in order to
avoid interaction with the operating system that may affect our trace.
*********************************************************************/
unsigned char IRQAllHandler(
	PKINTERRUPT InterruptObject,
	PVOID Extension
	)
{
	int IRQNum, IntGenMask;
	KIRQL OrigIrql;

	// Find out which IRQ caused us to be called.  We are sneaky and
	//  pass the IRQ number directly in the extension pointer
	//  variable.
	IRQNum = (int)Extension;

	// For the interrupt delay function, we don't do the
	//  normal processing.
	if(IRQNum == int_info.SerIrq) {
		HandleSerialIRQ();
		return 1;
	}

	// don't Trace while serial active
	if(int_info.SerIrq != 0)
		return 1;

	Trace8259("Enter", IRQNum, -1);
	IntGenMask = -1;

	// See if we need to generate any interrupts.
	while(int_info.IntGenTable[int_info.IntGenIdx] != -1) {

		++int_info.IntGenIdx;

		// Disable interrupts
		_asm pushfd
		_asm cli

		// Activate interrupt line
		IntGenMask =
				GenIRQ(int_info.IntGenTable[int_info.IntGenIdx-1]);
		Trace8259(" Gen", IRQNum, IntGenMask);

		// Enable interrupts
		_asm popfd
	}

	Trace8259("Leave", IRQNum, -1);
	return 1;
}

/*********************************************************************
  Set up an interrupt handler.  Take the desired IRQL as an argument.
If DesiredIRQL is -1, just use the default IRQL for the
"SynchronizeIRQL" parameter to IoConnectInterrupt().  Otherwise, set
sirql to DesiredIRQL.
*********************************************************************/
NTSTATUS SetupIRQ(
	int irq,			// IRQ number
	int DesiredIRQL,	// Desired IRQL
	unsigned char int_handler(PKINTERRUPT, PVOID),
	PKINTERRUPT *IntObj)
{
    ULONG       interruptVector;
    KIRQL       irql, sirql;
    KAFFINITY   affinity;
    NTSTATUS    status = STATUS_SUCCESS;

	//  When we call to "boost" an interrupt's priority, we first
	// have to disconnect it.
	if(*IntObj != NULL) {
		ResetDirectInt(irq);
		IoDisconnectInterrupt(IntObjTable[irq]);
	}

	interruptVector = HalGetInterruptVector(Isa, 0, irq, irq,
												&irql, &affinity);
	sirql = irql;

	if(DesiredIRQL != -1)
		sirql = DesiredIRQL;

	// Be sneaky, just stuff the IRQ number into the
	// "extension" parameter so that our "generic"
	// interrupt routine knows which interrupt occurred.
	// Luckily, kernel doesn't try to use it as a pointer.
	status = IoConnectInterrupt(IntObj,
								int_handler,
								(PVOID)irq,
								NULL,
                                interruptVector,
								irql, sirql,
                                Latched,
								FALSE,
								affinity,
                                FALSE);
	//
	// Set up the IDT to point to the assembler routine.
	// This must be done AFTER IoConnectInterrupt(), else NT won't
	// connect the interrupt.
	//
	if(status == STATUS_SUCCESS)
		SetDirectInt(interruptVector);
	return status;
}

/*********************************************************************
  Set up an interrupt handler with the default priority.
*********************************************************************/
NTSTATUS SetupIRQDefault(int irq)
{
	return SetupIRQ(irq, -1, IRQAllHandler, &IntObjTable[irq]);
}

/*********************************************************************
  Set up an interrupt handler with a priority boosted up to HIGH_LEVEL
(the highest available IRQL).
*********************************************************************/
NTSTATUS SetupIRQBoost(int irq)
{
	return SetupIRQ(irq, HIGH_LEVEL,
					IRQAllHandler, &IntObjTable[irq]);
}

/*********************************************************************
This routine checks out IRQs to see which ones may be used for
manually generating interrupts.  We take advantage of the behavior
of the edge/level setting register that causes an interrupt to
be generated on some IRQs when the edge/level mode is set to level.

It then allocates interrupt objects for all of the available IRQs and
submits them to NT so we can make sure to use only interrupts that no
one else is using.
*********************************************************************/
void SetupAvailableIRQs(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
	NTSTATUS status;
	PCM_RESOURCE_LIST AllocatedResources;
	PIO_RESOURCE_REQUIREMENTS_LIST RequestedResources;
	int ListSize;
	int imr, level;
	int active;
	int i, irq, mask;
	KIRQL OrigIrql;

	// save mask and edge/level registers
	imr = _inp(I8259_1_MASK);
	imr |= _inp(I8259_2_MASK) << 8;

	// Disable interrupts
	KeRaiseIrql(HIGH_LEVEL, &OrigIrql);

	// Save original edge/level status.
	level = _inp(I8259_1_EDGELVL) | (_inp(I8259_2_EDGELVL) << 8);

	// set up to read IRR
	_outp(I8259_1_CMD, 0x0a);
	_outp(I8259_2_CMD, 0x0a);

	// twiddle edge/level bits
	_outp(I8259_1_EDGELVL, 0xff);
	_outp(I8259_2_EDGELVL, 0xff);

	// See what interrupts were generated.
	active = _inp(I8259_1_CMD) | (_inp(I8259_2_CMD) << 8);

	// restore registers
	_outp(I8259_1_EDGELVL, level);
	_outp(I8259_2_EDGELVL, level >> 8);

	// Enable interrupts.
	KeLowerIrql(OrigIrql);

	// must be active AND available (i.e. masked)
	active &= imr;

	// don't generate ones that we know are off limits
	active &= OFF_LIMITS_IRQ_MASK;

	//
	// Let NT know what interrupts we will be using.  This stores the
	// interrupts we use in the registry so that conflicts with other
	// devices are avoided.  Also makes us show up in the NT
	// Diagnostics app so users can see which interrupts are okay
	// to use.
	//
	ListSize = sizeof(IO_RESOURCE_REQUIREMENTS_LIST) +
					15 * sizeof(IO_RESOURCE_DESCRIPTOR);
	RequestedResources = ExAllocatePool(PagedPool, ListSize);
	RtlZeroMemory(RequestedResources, ListSize);

	RequestedResources->InterfaceType = Isa;
	RequestedResources->AlternativeLists = 1;
	RequestedResources->List[0].Version = 1;
	RequestedResources->List[0].Revision = 1;

	for(irq=0, mask=1, i=0; irq < 16; ++irq, mask <<= 1)
		if(active & mask) {
			PIO_RESOURCE_DESCRIPTOR prd;

			prd = &RequestedResources->List[0].Descriptors[i++];
			prd->Option = IO_RESOURCE_PREFERRED;
			prd->Type = CmResourceTypeInterrupt;
			prd->ShareDisposition = CmResourceShareDriverExclusive;
			prd->Flags = CM_RESOURCE_INTERRUPT_LATCHED;
			prd->u.Interrupt.MinimumVector = irq;
			prd->u.Interrupt.MaximumVector = irq;
		}

	RequestedResources->ListSize = 
				sizeof(IO_RESOURCE_REQUIREMENTS_LIST) +
					i * sizeof(IO_RESOURCE_DESCRIPTOR);
	RequestedResources->List[0].Count = i;

	status = IoAssignResources(RegistryPath, NULL, DriverObject, 0,
						RequestedResources, &AllocatedResources);
	if(status != STATUS_SUCCESS)
		goto getout;

	for(i=0; i < 16; ++i)
		IntObjTable[i] = NULL;

	//
	// Only use those interrupts that NT says are free.
	//
	int_info.EnabledIRQMask = 0;
	for(i=0; i < (int)AllocatedResources->
							List[0].PartialResourceList.Count; ++i) {
		irq = AllocatedResources->List[0].PartialResourceList.
							PartialDescriptors[i].u.Interrupt.Level;
		SetupIRQDefault(irq);
		int_info.EnabledIRQMask |= 1 << irq;
	}

	ExFreePool(AllocatedResources);

getout:
	ExFreePool(RequestedResources);
}

/*********************************************************************
  Unload the driver.  Disconnects all interrupts.
*********************************************************************/
VOID DriverUnload(PDRIVER_OBJECT DriverObject)
{
	NTSTATUS status;
	PCM_RESOURCE_LIST AllocatedResources;
	WCHAR DOSNameBuffer[] = L"\\DosDevices\\" DEVICE_NAME_STRING;
	UNICODE_STRING uniDOSString;
	int i;

	for(i=0; i < 16; ++i)
		if(IntObjTable[i] != NULL) {
			ResetDirectInt(i);
			IoDisconnectInterrupt(IntObjTable[i]);
		}

	status = IoAssignResources(&DriverRegistryPath, 0, DriverObject,
								0, 0, &AllocatedResources);

	if(status == STATUS_SUCCESS && AllocatedResources != NULL)
		ExFreePool(AllocatedResources);
	ExFreePool(DriverRegistryPath.Buffer);

	RtlInitUnicodeString(&uniDOSString, DOSNameBuffer);
	IoDeleteSymbolicLink (&uniDOSString);
	IoDeleteDevice(DriverObject->DeviceObject);
}

/*********************************************************************
 Generic service handler.  Does not too much.
*********************************************************************/
NTSTATUS
CreateDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return STATUS_SUCCESS;
}

/*********************************************************************
 Must release app's mapping of int_info before the thread terminates.
Also, "un-boost" any boosted interrupt.
*********************************************************************/
NTSTATUS
DriverClose(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
	UnMapBuffer();

	if(BoostIRQ) {
		SetupIRQDefault(BoostIRQ);
		BoostIRQ = 0;
	}

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return STATUS_SUCCESS;
}

/*********************************************************************
  Handle our one and only IOCTL call.  Map int_info structure into
caller's address space.
*********************************************************************/
NTSTATUS
DeviceControl(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PIO_STACK_LOCATION		irpSp;
    NTSTATUS				status;
	char					*buf;
	INT_INFO				*info_ptr;

	status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    irpSp = IoGetCurrentIrpStackLocation(Irp);
	buf = Irp->AssociatedIrp.SystemBuffer;

	BoostIRQ = 0;
	if(irpSp->Parameters.DeviceIoControl.InputBufferLength
												== sizeof(int)) {
		BoostIRQ = *(int *)buf;
		SetupIRQBoost(BoostIRQ);
	}

	if(irpSp->Parameters.DeviceIoControl.OutputBufferLength
					< sizeof(INT_INFO *))
		status = STATUS_BUFFER_TOO_SMALL;
	else {
		info_ptr = MapBuffer(&int_info, sizeof(INT_INFO));
		memcpy(buf, &info_ptr, sizeof(INT_INFO *));
		Irp->IoStatus.Information = sizeof(INT_INFO *);
	}

    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return STATUS_SUCCESS;
}

/*********************************************************************
  Entry routine.  Set everything up.
*********************************************************************/
NTSTATUS DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
	PDEVICE_OBJECT deviceObject;
	NTSTATUS status;
	WCHAR NameBuffer[] = L"\\Device\\" DEVICE_NAME_STRING;
	WCHAR DOSNameBuffer[] = L"\\DosDevices\\" DEVICE_NAME_STRING;
	UNICODE_STRING uniNameString, uniDOSString;

	// Save for Unload routine.
	DriverRegistryPath.Buffer =
		ExAllocatePool(PagedPool, RegistryPath->MaximumLength);
	DriverRegistryPath.MaximumLength = RegistryPath->MaximumLength;
	RtlCopyUnicodeString(&DriverRegistryPath, RegistryPath);

	//
	//  Set up device driver name and device object.
	//
	RtlInitUnicodeString(&uniNameString, NameBuffer);
	RtlInitUnicodeString(&uniDOSString, DOSNameBuffer);

	status = IoCreateDevice(DriverObject, sizeof(PVOID),
					&uniNameString,
					FILE_DEVICE_UNKNOWN,
					0, FALSE, &deviceObject);

    if(!NT_SUCCESS(status))
		return status;

	status = IoCreateSymbolicLink (&uniDOSString, &uniNameString);

	deviceObject->Flags |= DO_BUFFERED_IO;

    //
    // Initialize the Driver Object with driver's entry points
    //
    DriverObject->MajorFunction[IRP_MJ_CREATE] = CreateDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = DriverClose;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
													 DeviceControl;
	DriverObject->DriverUnload = DriverUnload;

	SetupAvailableIRQs(DriverObject, RegistryPath);

    return STATUS_SUCCESS;
}

