/*********************************************************************
Author:       Dale Roberts
File:         GENIRQ.C

Purpose:

Console application portion of the GENIRQ/DOIRQ team.  Together this
application and the DOIRQ kernel-mode driver allow you to generate
overlapped interrupts and monitor NT's handling of the interrupt
service routines.  It also sets up the a serial port to generate
periodic interrupts for interrupt delay timing purposes.
*********************************************************************/
#define WIN32

#include <stdio.h>
#include <windows.h>
#include <winioctl.h>
#include <conio.h>
#include "intinfo.h"

//
// Clock rate of system timer 0.
// Used to calibrate the RDTSC (Read CPU Time Stamp Counter)
// instruction.
//
#define TIMER0_RATE		(14318180.0 / 12.0)

//
//  Serial port and IRQ number used for the interrupt timing routine.
//
#define	SER_PORT	0x2f8
#define SER_IRQ		3

/*********************************************************************
  Use the "undocumented" Pentium RDTSC (Read TimeStamp Counter)
instruction to get a very high resolution timestamp.  The assembler
instruction returns the value of a counter inside the Pentium that is
updated on every CPU clock tick (i.e., about 133,000,000 times per
second for a 133MHz processor).

  Since the instruction is not known by the inline assembler, we must
directly emit its two byte opcode.

  The RDTSC instruction returns the 64-bit counter in the EDX:EAX
(upper 32 bits : lower 32 bits) registers.  Luckily, these are the
same two registers that MS C uses to return an _int64 (a 64-bit
integer). So all the routine needs to do is issue the assembler
instruction and leave the results in the EDX:EAX register.  There is
no need to explicitely "return" the data.
*********************************************************************/
_int64 RDTSC(void)
{
	_asm _emit 0x0f		// RDTSC Read Timestamp Counter instruction
	_asm _emit 0x31		// (inline assembler doesn't know this one)
						// Puts time stamp in EDX:EAX

	// Microsoft C returns an _int64 in EDX:EAX.  Since the counter
	// value is already stored there, we don't have to do anyting
	// in order to return it to the caller.
}

/*********************************************************************
  Read the value of the system counter/timer.  This counter is used to
generate periodic interrupts in NT for preemptive multitasking.  We
simply read it here, and do not modify it or change its behavior in
any way.
*********************************************************************/
int ReadTimer0(void)
{
	int lo;

	// counter 0 read-back
	_outp(0x43, 0);
	lo = _inp(0x40);
	return lo | (_inp(0x40) << 8);
}

/*********************************************************************
  Do a rough measure of CPU clock speed.

  To do this we need a known timebase.  We use the system timer 0. 
This is updated at a known frequency of 14,318,180 / 12 per second on
all AT compatible systems.  We count through 5000 ticks of this clock
(we know how long this takes) and see how many CPU clock ticks occur
during this time.

  We return the number of CPU clock ticks per second.

  We are not very careful about subtracting out any overhead in the
timing process itself, so this value is pretty rough, but plenty
accurate for our needs.
*********************************************************************/
int CalcCPUSpeed(void)
{
	_int64 StartStamp, EndStamp;
	int startcnt, endcnt, CPUSpeed;
	int imr0;

	// wait for count > 5000, since we want to count down
	//  for 5000 ticks.
	while(ReadTimer0() < 5100)
		;

	// disable interrupts
	imr0 = _inp(I8259_1_MASK);
	_outp(I8259_1_MASK, 0xff);

	// Get the current count and start counting just below that.
	startcnt = ReadTimer0() - 2;
	endcnt = startcnt - 5000;
	while(ReadTimer0() > startcnt)
		;

	// read initial time stamp
	StartStamp = RDTSC();

	// wait for 5000 TIMER0 ticks
	while(ReadTimer0() > endcnt)
		;

	// read end time stamp
	EndStamp = RDTSC();

	// enable interrupts
	_outp(I8259_1_MASK, imr0);

	CPUSpeed = (int)((double)(EndStamp-StartStamp)
				/ (5000.0 / TIMER0_RATE));

	return CPUSpeed;
}

/*********************************************************************
  Generate one or more IRQs using the devious "twiddle the edge/level
bit" trick.
*********************************************************************/
void GenIRQ(INT_INFO *int_info, int IRQMask)
{
	int imr0;

	// Only trigger those interrupts that the driver
	//  successfully claimed.
	IRQMask &= int_info->EnabledIRQMask;

	imr0 = _inp(I8259_1_MASK);

	// Yes, port 0x21 could change in between here, but we'll cross
	// our fingers and hope it doesn't.

	// disable interrupts
	_outp(I8259_1_MASK, 0xff);

	_outp(I8259_1_EDGELVL, _inp(I8259_1_EDGELVL) | IRQMask);
	_outp(I8259_2_EDGELVL, _inp(I8259_2_EDGELVL) | (IRQMask >> 8));

	// enable interrupts
	_outp(I8259_1_MASK, imr0);
}

/*********************************************************************
  Set up the serial port to generate interrupts.  This is standard
serial port stuff, so no detailed information is given here.  We
enable serial port interrupt generation for the "output buffer empty"
situation, then put two bytes into the serial data output register. 
When the first byte goes out of the transmit shift register, and the
last byte is read out of the transmit holding register, an interrupt
is generated. The interrupt routine puts another byte into the output
register to keep the process going indefinately.
*********************************************************************/
void StartSerInts(void)
{
	_outp(SER_PORT+1, 0x00);	// disable serial ints
	_inp(SER_PORT+2);			// clear out interrupts
	_inp(SER_PORT+6);			// read status to clear possible irq

	_outp(SER_PORT+3, 0x80);	// enable access to divisor regs
	_outp(SER_PORT, 2);			// set to 5760 char/sec
	_outp(SER_PORT+1,0);
	_outp(SER_PORT+3, 0x3);		// 8 bits, 1 stop bit

	_outp(SER_PORT+4, 0x08);	// master interrupt enable
	_outp(SER_PORT+1, 0x02);	// enable transmit interrupt
	_outp(SER_PORT, 0x00);		// put two bytes out; one for shift
	_outp(SER_PORT, 0x00);		//  reg; one for holding reg
}

/*********************************************************************
  Stop the serial interrupts.
*********************************************************************/
void EndSerInts(void)
{
	_outp(SER_PORT+4, 0x00);	// master interrupt disable
	_outp(SER_PORT+1, 0x00);	// disable transmit interrupts
	_inp(SER_PORT+2);			// clear out interrupts
	_inp(SER_PORT+6);			// read status to clear possible irq
}

/*********************************************************************
  Start up the serial interrupts and display the realtime information.
The int_info structure is updated directly by the interrupt routine.
We display the current and maximum inter-interrupt time, as well as
a count of interrupts that are significantly delayed.
*********************************************************************/
void DoSerialPort(INT_INFO *int_info, double TicksPerUSec)
{
	char ch;

	printf("SPACE bar clears out statistics, other keys quit.\n\n");

	int_info->SerPort = SER_PORT;
	int_info->SerIrq = SER_IRQ;
	int_info->MaxSerIRQInterval = 0;
	int_info->LastSerTimestamp = RDTSC();
	int_info->NumDelayedInts = 0;
	int_info->NumDelayedInts2 = 0;

	StartSerInts();

	ch = ' ';
	while(ch == ' ') {
		if(kbhit()) {
			ch = getch();
			if(ch == ' ') {
				int_info->MaxSerIRQInterval = 0;
				int_info->NumDelayedInts = 0;
				int_info->NumDelayedInts2 = 0;
			}
		}
		printf("Cur %8.1fuS   Max %8.1fuS"
					"   #>400uS %7d   #>1000uS %7d\r",
				(double)int_info->CurSerIRQInterval / TicksPerUSec,
				(double)int_info->MaxSerIRQInterval	/ TicksPerUSec,
				int_info->NumDelayedInts,
				int_info->NumDelayedInts2);
	}

	EndSerInts();
	int_info->SerPort = 0;
	int_info->SerIrq = 0;

	while(kbhit())
		getch();
}

/*********************************************************************
  Dump the table of trace information generated from the overlapped
interrupts.
*********************************************************************/
DumpTrace(INT_INFO *int_info, double TicksPerUSec, int BoostIRQ)
{
	INT_TRACE *tp;
	_int64 LastStamp;
	char IRQMaskStr[10];
	int i;

	LastStamp = int_info->Trace[0].TimeStamp;

	printf("           IRQ   IRR  ISR  IMR IF   uSec\n");

	for(i=0; i < int_info->TraceIdx; ++i) {
		tp = &(int_info->Trace[i]);
		sprintf(IRQMaskStr, "%04x", tp->GenIRQMask);
		printf("%-5.5s %4.4s #%2d%c %04x %04x %04x %d %7.2f\n",
			tp->msg,
			tp->GenIRQMask == 0xffff ? "" : IRQMaskStr,
			tp->IntNum,
			tp->IntNum == BoostIRQ ? '!' : ' ',
			tp->IRR, tp->ISR, tp->IMR,
			tp->IFlag,
			(double)(tp->TimeStamp - LastStamp) / TicksPerUSec);
		LastStamp = tp->TimeStamp;
	}
}

/*********************************************************************
  Add the interrupt entry into the specified interrupt level's queue.
This queue is read by the interrupt routine to generate subsequent
interrupts.
*********************************************************************/
void AddInt(INT_INFO *int_info, int Mask)
{
	// eliminate off limits IRQs
	Mask &= OFF_LIMITS_IRQ_MASK;

	if(int_info->IntGenIdx >= (INT_GEN_TABLE_SIZE - 1))
		return;

	int_info->IntGenTable[int_info->IntGenIdx] = Mask;
	++int_info->IntGenIdx;
	int_info->IntGenTable[int_info->IntGenIdx] = -1;
}

/*********************************************************************
  main()

 - Get access to I/O ports using GIVEIO.SYS driver.
 - Get access to the DOIRQ driver.
 - Check command line args to see if we need a priority boost.
 - Map int_info structure into our memory space, and possibly boost
   an IRQ.
 - Compute CPU speed.
 - Read IRQ numbers from command line and build queue of ints to
   generate.
 - Generate initial interrupt and let DOIRQ do the rest.
 - Display trace table.
 - Do serial port interrupt delay timing.
*********************************************************************/
int main(int argc, char *argv[])
{
	HANDLE h;
	DWORD err;
	INT_INFO *int_info;
	int i, nchar;
	double TicksPerUSec;
	int BoostIRQ;

	//
	// Use GIVEIO.SYS to get access to I/O ports.
	//
    h = CreateFile("\\\\.\\giveio", GENERIC_READ, 0, NULL,
					OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if(h == INVALID_HANDLE_VALUE) {
        printf("Couldn't access giveio device\n");
        return -1;
    }
	CloseHandle(h);

	//
	// Get handle to our DOIRQ.SYS driver
	//
	h = CreateFile("\\\\.\\doirq", GENERIC_READ, 0, NULL,
				OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(h == INVALID_HANDLE_VALUE) {
		err = GetLastError();
		printf("Could not open DOIRQ driver.  Error: %ld\n", err);
		return -1;
	}

	// See if we need an IRQ priority boost.  Look for '!'.
	for(BoostIRQ=0, i=1; i < argc; ++i)
		if(strchr(argv[i], '!')) {
			BoostIRQ = atoi(argv[i]);
			break;
		}

	// Map driver's int_info structure into our memory space.
	DeviceIoControl(h, IOCTL_TESTDRV_GO, &BoostIRQ, sizeof(BoostIRQ),
					&int_info, sizeof(INT_INFO *), &nchar, 0);

	// Compute CPU speed.
	TicksPerUSec = CalcCPUSpeed() / 1000000.0;
	int_info->CPUTicksPerUSec = (int)TicksPerUSec;

	printf("CPU Speed %.1f MHz\n", TicksPerUSec);
	printf("Using Serial port 0x%x, Serial IRQ %d\n",
			SER_PORT, SER_IRQ);

	// Show available IRQs as determined by DOIRQ.
	printf("Available IRQ mask: 0x%04x:", int_info->EnabledIRQMask);
	for(i=0; i < 16; ++i)
		if(int_info->EnabledIRQMask & (1 << i))
			printf(" IRQ%d", i);
	printf("\n");

	// Initialize trace table index.
	int_info->TraceIdx = 0;

	// Initialize interrupt generation queue.
	int_info->IntGenIdx = 0;
	int_info->IntGenTable[0] = -1;

	// Read interrupt numbers from command line.
	while(--argc > 0)
		int_info->IntGenTable[int_info->IntGenIdx++]
									= 1 << atoi(*(++argv));

	// Terminate IRQ queue with -1.
	int_info->IntGenTable[int_info->IntGenIdx] = -1;
	int_info->IntGenIdx = 0;

	// If there are interrupts to be generated, then do it!
	if(int_info->IntGenTable[0] != -1) {
		++int_info->IntGenIdx;
		GenIRQ(int_info, int_info->IntGenTable[0]);
	}

	// Show the interrupt traces to the user.
	printf("\n");
	DumpTrace(int_info, TicksPerUSec, BoostIRQ);
	printf("\n");

	//
	// Move on to the delay timing portion of our show.
	//
	DoSerialPort(int_info, TicksPerUSec);

closeit:
	if(h != INVALID_HANDLE_VALUE)
		CloseHandle(h);
	return 0;
}
