// --------- ptl.h
#ifndef PTL_H
#define PTL_H
// ==========================================================
// PTL Version 2,
// a persistent template container library
// based on the C++ Standard Template Library (STL).
// For documentation on the construction of this library and
// its use, read the C Programming Column by Al Stevens in 
// the December 97 and April 98 issues of Dr. Dobb's Journal.
// Please send all comments to astevens@ddj.com
// ==========================================================
#include <vector>
#include <deque>
#include <list>
#include <stack>
#include <queue>
#include <map>
#include <set>
#include <string>
#include <ptlstore.h>

namespace PersistentTemplateLibrary	{

// ========================================================
//         PTLcntr class
//   abstract base persistent container class
//   derived from STL templatized container
// ========================================================
template <class C>
class PTLcntr : public C
{
	// ---- cannot be copied
    PTLcntr(const PTLcntr&) { }
	PTLcntr& operator=(const PTLcntr& pc)
		{ return *this; }
protected:
    PTLStore cs;
	bool changed;
	explicit PTLcntr(const std::string& name)
	{
		// --- indicates container has been modified
		changed = false;
		// --- open for input; derived ctors read and close
	    cs.open(name + ".ptl");
	}
    virtual ~PTLcntr()
    	{ }
	virtual void save() = 0;
public:
	void setchanged()			// mark container to be rewritten
		{ changed = true; }
	void suppress_save()		// unmark container-rewritten flag
		{ changed = false; }
	void save_now()
		{ save(); }				// save now 
	bool ischanged() const
		{ return changed; }
};
// ========================================================
//                  PTLseq
//   base class for PTLvector, PTLdeque, and PTLlist
// ========================================================
template <class T, class C>
class PTLseq : public PTLcntr<C>
{
	typedef PTLcntr<C> B;
	// ---- save the container's objects to the ptl store
	void save()
	{
	    cs.open(size());
	    for (iterator iter = begin(); iter != end(); iter++)
	   	    cs << *iter;
		changed = false;
	    cs.close();
	}
protected:
    explicit PTLseq(const std::string& name) : B(name)
	{
		// --- cs is opened and objcount is loaded by PTLcntr ctor
    	for (int i = 0; i < cs.count(); i++)    {
	        T t;
	        cs >> t;
	        push_back(t);
	    }
	    cs.close();
	}
    ~PTLseq()
		{ if (changed) save(); }
public:
	// --- override STL modification functions to mark container as changed
	C::reference front()
		{ changed = true; return B::front(); }
	C::reference back()
		{ changed = true; return B::back(); }
	void push_back(const T& x)
		{ changed = true; B::push_back(x); }
	void pop_back()
		{ changed = true; B::pop_back(); }

	C::iterator insert(C::iterator position, const T& x) // = T())  // VC++ 5.0 bug
		{ changed = true; return B::insert(position, x); }
	void insert(C::iterator position, C::size_type n, const T& x)
		{ changed = true; B::insert(position, n, x); }

	template <class InputIterator>
	void insert(C::iterator position, InputIterator first, InputIterator last)
		{ changed = true; B::insert(position, first, last); }

	C::iterator erase(C::iterator position)
		{ changed = true; return B::erase(position); }
	C::iterator erase(C::iterator first, C::iterator last)
		{ changed = true; return B::erase(first, last); }
	void clear()
		{ changed = true; B::clear(); }
};
// ========================================================
//                    PTLassoc
//     base class for PTLsets and PTLmaps
// ========================================================
template <class C>
class PTLassoc : public PTLcntr<C>
{
	typedef PTLcntr<C> B;
protected:
    explicit PTLassoc(const std::string& name) : B(name) { }
public:
	// --- override STL modification functions to mark container as changed
	C::iterator insert(C::iterator position, const C::value_type& x)
		{ changed = true; return B::insert(position, x); }
	template <class InputIterator>
	void insert(C::iterator position, InputIterator first, InputIterator last)
		{ changed = true; B::insert(position, first, last); }

	void erase(C::iterator position)
		{ changed = true; B::erase(position); }
	C::size_type erase(const C::key_type& x)
		{ changed = true; return B::erase(x); }
	void erase(C::iterator first, C::iterator last)
		{ changed = true; B::erase(first, last); }
	void clear()
		{ changed = true; B::clear(); }
};
// ========================================================
//                    PTLsets
//     base class for PTLset and PTLmultiset
// ========================================================
template <class T, class C>
class PTLsets : public PTLassoc<C>
{
	void save()
	{
	    cs.open(size());
	    for (iterator iter = begin(); iter != end(); iter++)
	   	    cs << *iter;
		changed = false;
	    cs.close();
	}
protected:
    explicit PTLsets(const std::string& name) : PTLassoc<C>(name)
	{
	    for (int i = 0; i < cs.count(); i++)    {
	        T t;
	        cs >> t;
			C::insert(t);
	    }
	    cs.close();
	}
    virtual ~PTLsets()
		{ if (changed) save(); }
};
// ========================================================
//                    PTLmaps
//     base class for PTLmap and PTLmultimap
// ========================================================
template <class K, class T, class C>
class PTLmaps : public PTLassoc<C>
{
	void save()
	{
	    cs.open(size());
	    for (iterator iter = begin(); iter != end(); iter++)    {
	   	    cs << iter->first;
	       	cs << iter->second;
	    }
		changed = false;
	    cs.close();
	}
protected:
    explicit PTLmaps(const std::string& name) : PTLassoc<C>(name)
	{
    	for (int i = 0; i < cs.count(); i++)    {
        	K k;
	        cs >> k;
    	    T t;
        	cs >> t;
			C::insert(std::pair<K,T>(k, t));
    	}
	    cs.close();
	}
    virtual ~PTLmaps()
		{ if (changed) save(); }
};
// =============================================================
//
//           PTL public interface
//
// template parameter notation:
//   T = type being contained
//   K = key value for sets and maps
//   C = STL container being used by container adaptor
//   P = predicate function
//
// =============================================================
//                 PTLvector
// =============================================================
template <class T, class A = std::allocator<T> >
class PTLvector : public PTLseq<T, std::vector<T,A> >	{
	typedef std::vector<T,A> C;
	typedef PTLseq<T, C> B;
public:
	explicit PTLvector(const std::string& name) : B(name)
		{ }
	// --- override STL modification functions to mark container as changed
	C::reference operator[](C::size_type n)
		{ changed = true; return B::operator[](n); }
	C::reference at(C::size_type n)
		{ changed = true; return B::at(n); }
};
// =============================================================
//                 PTLdeque
// =============================================================
template <class T, class A = std::allocator<T> >
class PTLdeque : public PTLseq<T, std::deque<T,A> >	{
	typedef std::deque<T,A> C;
	typedef PTLseq<T, C> B;
public:
	explicit PTLdeque(const std::string& name) : B(name)
		{ }
	// --- override STL modification functions to mark container as changed
	C::reference operator[](C::size_type n)
		{ changed = true; return B::operator[](n); }
	C::reference at(C::size_type n)
		{ changed = true; return B::at(n); }

	void push_front(const T& x)
		{ changed = true; B::push_front(x); }
	void pop_front()
		{ changed = true; B::pop_front(); }
};
// =============================================================
//                 PTLlist
// =============================================================
template <class T, class A = std::allocator<T> >
class PTLlist : public PTLseq<T, std::list<T,A> >	{
	typedef std::list<T,A> C;
	typedef PTLseq<T, C> B;
public:
	explicit PTLlist(const std::string& name) : B(name)
		{ }
	// --- override STL modification functions to mark container as changed
	void push_front(const T& x)
		{ changed = true; B::push_front(x); }
	void pop_front()
		{ changed = true; B::pop_front(); }
};
// =============================================================
//                 PTLstack
// =============================================================
template <class T, class C = std::deque<T> >
class PTLstack : public PTLcntr<std::stack<T,C> >
{
	typedef std::stack<T,C> aC;
	typedef PTLcntr<aC> B;
	void save()
	{
    	cs.open(size());
	    aC stack;
   		while (!empty())    {
       		stack.push(top());
	        pop();
    	}
	   	while (!stack.empty())    {
    	   	cs << stack.top();
        	stack.pop();
		}
		changed = false;
	    cs.close();
	}
public:
    explicit PTLstack(const std::string& name) : B(name)
	{
    	for (int i = 0; i < cs.count(); i++)    {
        	T t;
	        cs >> t;
    	    push(t);
	    }
	    cs.close();
	}
    ~PTLstack()
		{ if (changed) save(); }

	// --- override STL modification functions to mark container as changed
	aC::value_type& top()
		{ changed = true; return B::top(); }
	void push(const aC::value_type& x)
		{ changed = true; B::push(x); }
	void pop()
		{ changed = true; B::pop(); }
};
// =============================================================
//                 PTLqueue
// =============================================================
template <class T, class C = std::deque<T> >
class PTLqueue : public PTLcntr<std::queue<T,C> >
{
	typedef std::queue<T,C> aC;
	typedef PTLcntr<aC> B;
	void save()
	{
    	cs.open(size());
	    while (!empty())    {
   		    cs << front();
       		pop();
		}
		changed = false;
	    cs.close();
	}
public:
    explicit PTLqueue(const std::string& name) : B(name)
	{
    	for (int i = 0; i < cs.count(); i++)    {
        	T t;
	        cs >> t;
    	    push(t);
	    }
	    cs.close();
	}
    ~PTLqueue()
		{ if (changed) save(); }
	// --- override STL modification functions to mark container as changed
	aC::value_type& front()
		{ changed = true; return B::front(); }
	std::queue<T, C>::value_type& back()
		{ changed = true; return B::back(); }
	void push(const aC::value_type& x)
		{ changed = true; B::push(x); }
	void pop()
		{ changed = true; B::pop(); }
};
// =============================================================
//                 PTLpriority_queue
// =============================================================
template <class T, class C=std::vector<T>, class P=std::less<C::value_type> >
class PTLpriority_queue : public PTLcntr<std::priority_queue<T,C,P> >
{
	typedef std::priority_queue<T,C,P> aC;
	typedef PTLcntr<aC> B;
	void save()
	{
	    cs.open(size());
	    while (!empty())    {
	   	    cs << top();
	       	pop();
		}
		changed = false;
	    cs.close();
	}
public:
    explicit PTLpriority_queue(const std::string& name) : B(name)
	{
		for (int i = 0; i < cs.count(); i++)    {
	    	T t;
		    cs >> t;
			push(t);
	    }
    	cs.close();
	}
    virtual ~PTLpriority_queue()
		{ if (changed) save(); }

	// --- override STL modification functions to mark container as changed
	void push(const aC::value_type& x)
		{ changed = true; B::push(x); }
	void pop()
		{ changed = true; B::pop(); }
};
// =============================================================
//                    PTLset
// =============================================================
template <class K, class P = std::less<K>, class A = std::allocator<K> >
class PTLset : public PTLsets<K, std::set<K,P,A> >	{
	typedef std::set<K,P,A> C;
	typedef PTLsets<K, C> B;
public:
	explicit PTLset(const std::string& name) : B(name)
		{ }
	std::pair<C::iterator,bool> insert(const C::value_type& x)
		{ changed = true; return C::insert(x); }
};
// =============================================================
//                    PTLmultiset
// =============================================================
template <class K, class P = std::less<K>, class A = std::allocator<K> >
class PTLmultiset : public PTLsets<K, std::multiset<K,P,A> >	{
	typedef std::multiset<K,P,A> C;
	typedef PTLsets<K, C> B;
public:
	explicit PTLmultiset(const std::string& name) : B(name)
		{ }
	C::iterator insert(const C::value_type& x)
		{ changed = true; return C::insert(x); }
};
// =============================================================
//                    PTLmap
// =============================================================
template <class K, class T, class P = std::less<K>, class A = std::allocator<T> >
class PTLmap : public PTLmaps<K, T, std::map<K,T,P,A> >	{
	typedef std::map<K,T,P,A> C;
	typedef PTLmaps<K,T,C> B;
public:
	explicit PTLmap(const std::string& name) : B(name)
		{ }
	// --- override STL modification functions to mark container as changed
	A::reference operator[](const C::key_type& k)	// s/b C::reference (VC++ STL bug)
		{ changed = true; return C::operator[](k); }
	std::pair<C::iterator,bool> insert(const C::value_type& x)
		{ changed = true; return C::insert(x); }
};
// =============================================================
//                    PTLmultimap
// =============================================================
template <class K, class T, class P = std::less<K>, class A = std::allocator<T> >
class PTLmultimap : public PTLmaps<K,T,std::multimap<K,T,P,A> >  {
	typedef std::multimap<K,T,P,A> C;
	typedef PTLmaps<K,T,C> B;
public:
	explicit PTLmultimap(const std::string& name) : B(name)
		{ }
	// --- override STL modification functions to mark container as changed
	C::iterator insert(const C::value_type& x)
		{ changed = true; return C::insert(x); }
};

} // namespace PersistentTemplateLibrary

#endif // PTL_H
